/***************************************************************************
 * jtalk -- talk-like Jabber client
 * (C) Nicolas George - 2006
 * jtalk_rl.c: readline-based front-end
 ***************************************************************************/

/*
  This file is part of jtalk.

  jtalk is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  jtalk is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with jtalk; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <locale.h>
#include <langinfo.h>
#include <iconv.h>
#include <poll.h>
#include <signal.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <readline/readline.h>
#include <readline/history.h>

#include "frontend_common.h"

/*
   esc 7 -> esc 8
   esc s ? -> esc u ?
*/

#define SAVE_CURSOR "\0337"
#define RESTORE_CURSOR_AND_CLEAR "\0338\033[J"

#define RL_PROMPT_START_IGNORE_S "\001"
#define RL_PROMPT_END_IGNORE_S "\002"

#define SAVE_CURSOR_PROMPT RL_PROMPT_START_IGNORE_S SAVE_CURSOR \
    RL_PROMPT_END_IGNORE_S

static int sock;
static unsigned char *buffer;
static size_t buffer_size, buffer_have;

static int iconv_do;
static iconv_t iconv_to;
static iconv_t iconv_from;
static char *iconv_buffer = NULL;
static size_t iconv_buffer_size = 0;

static int in_is_tty, out_is_tty, on_tty;
static int in_enabled;

static void
enlarge_iconv_buffer(char **b, size_t *bs)
{
    size_t o;

    o = *b - iconv_buffer;
    iconv_buffer_size = iconv_buffer_size == 0 ? 1024 : iconv_buffer_size * 2;
    if((iconv_buffer = realloc(iconv_buffer, iconv_buffer_size)) == NULL) {
	perror("malloc");
	exit(1);
    }
    *b = iconv_buffer + o;
    *bs = iconv_buffer_size - o;
}

static char *
convert_line(iconv_t c, const char *l)
{
    size_t in_len, out_len;
    char *in, *out;
    int must_enlarge = 0;

    iconv(c, NULL, NULL, NULL, NULL);
    in = (char *)l;
    in_len = strlen(in);
    out = iconv_buffer;
    out_len = iconv_buffer_size;
    while(1) {
	if(out_len < 64 || must_enlarge)
	    enlarge_iconv_buffer(&out, &out_len);
	iconv(c, &in, &in_len, &out, &out_len);
	if(in_len == 0)
	    break;
	switch(errno) {
	    case E2BIG:
		must_enlarge = 1;
		break;
	    case EILSEQ:
	    case EINVAL:
		if(out_len < 1)
		    enlarge_iconv_buffer(&out, &out_len);
		in++;
		in_len--;
		*(out++) = '?';
		out_len--;
		break;
	    default:
		perror("iconv");
		exit(1);
	}
    }
    if(out_len < 1)
	enlarge_iconv_buffer(&out, &out_len);
    *out = 0;
    return(iconv_buffer);
}

static void
display_message(const char *t)
{
    printf("%s%s%s\n", on_tty ? RESTORE_CURSOR_AND_CLEAR : "", t,
	on_tty ? "\033[m" : "");
    fflush(stdout);
    rl_forced_update_display();
}

static void
send_to_jtalkd(const char *t)
{
    struct sigaction ign_sigpipe, old_sigaction;
    size_t l;
    ssize_t w;

    ign_sigpipe.sa_handler = SIG_IGN;
    sigemptyset(&ign_sigpipe.sa_mask);
    ign_sigpipe.sa_flags = 0;
    sigaction(SIGPIPE, &ign_sigpipe, &old_sigaction);
    l = strlen(t) + 1;
    while(l > 0) {
	if((w = write(sock, t, l)) < 0) {
	    if(errno == EINTR)
		continue;
	    /* with a well-behaved jtalkd (no half-shutdown on the socket),
	     * the closure will be reported by the poll(); on the other
	     * hand, we do not want a sigpipe while sending the final /argv
	     * if the last argument was /quit. */
	    if(errno == EPIPE)
		break;
	    perror("write");
	    exit(1);
	}
	l -= w;
	t += w;
    }
    sigaction(SIGPIPE, &old_sigaction, NULL);
}

static void
have_keyboard_line(char *line)
{
    char *good_line;

    if(in_is_tty && out_is_tty)
	printf(RESTORE_CURSOR_AND_CLEAR);
    if(line == NULL) {
	if(!in_is_tty)
	    in_enabled = 0;
	return;
    }
    add_history(line);
    good_line = line;
    if(iconv_do)
	good_line = convert_line(iconv_from, good_line);
    send_to_jtalkd(good_line);
    free(line);
}

static void
have_network_line(char *line)
{
    if(iconv_do)
	line = convert_line(iconv_to, line);
    display_message(line);
}

static void
set_socket_path(char *p, size_t lp)
{
    char *e;

    if((e = getenv("JTALK_BASE")) != NULL) {
	snprintf(p, lp, "%s/socket/s", e);
    } else {
	if((e = getenv("HOME")) == NULL) {
	    fprintf(stderr, "No $HOME nor $JTALK_BASE.\n");
	    exit(1);
	}
	snprintf(p, lp, "%s/.jtalk/socket/s", e);
    }
}

static void
connect_jtalkd(void)
{
    struct sockaddr_un addr;

    if((sock = socket(AF_UNIX, SOCK_STREAM, 0)) < 0) {
	perror("socket");
	exit(1);
    }
    memset(&addr, 0, sizeof(addr));
    addr.sun_family = AF_UNIX;
    set_socket_path(addr.sun_path, sizeof(addr.sun_path));
    if(connect(sock, (struct sockaddr *)&addr, sizeof(addr)) < 0) {
	perror("connect");
	exit(1);
    }
}

static void
read_from_jtalkd(void)
{
    ssize_t l;
    unsigned char *end;

    if(buffer_size - buffer_have < 1024) {
	buffer_size = buffer_size == 0 ? 4096 : buffer_size * 2;
	if((buffer = realloc(buffer, buffer_size)) == NULL) {
	    perror("malloc");
	    exit(1);
	}
    }
    if((l = read(sock, buffer + buffer_have, buffer_size - buffer_have)) < 0) {
	if(errno == EINTR)
	    return;
	perror("read");
	exit(1);
    }
    if(l == 0)
	exit(0);
    buffer_have += l;
    /* TODO: make that a little bit faster: do not re-search what is already
     * here, do not memmove for all lines (not really important, most of the
     * time I will have exactly one line at once in the buffer) */
    while((end = memchr(buffer, 0, buffer_have)) != NULL) {
	have_network_line(buffer);
	end++;
	buffer_have -= end - buffer;
	memmove(buffer, end, buffer_have);
    }
}

static void
init_iconv(void)
{
    char *codeset;

    if((codeset = nl_langinfo(CODESET)) == NULL) {
	perror("nl_langinfo(CODESET)");
	exit(1);
    }
    if(strcmp(codeset, "UTF-8") == 0) {
	iconv_do = 0;
    } else {
	char cst[strlen(codeset) + 11];

	sprintf(cst, "%s//TRANSLIT", codeset);
	if((iconv_from = iconv_open("UTF-8", codeset)) == (iconv_t)-1 ||
	    ((iconv_to = iconv_open(cst, "UTF-8")) == (iconv_t)-1 &&
	    (iconv_to = iconv_open(codeset, "UTF-8")) == (iconv_t)-1)) {
	    perror("iconv_open");
	    exit(1);
	}
	iconv_do = 1;
    }
}

static void
send_argv(int argc, char **argv)
{
    size_t l;

    while(argc > 0) {
	l = strlen(*argv);
	char b[7 + l];
	sprintf(b, "/argv %s", *argv);
	send_to_jtalkd(b);
	argv++;
	argc--;
    }
    send_to_jtalkd("/argv");
}

int
main(int argc, char **argv)
{
    struct pollfd fds[2];

    jtalk_frontend_select(argv, argc);
    in_is_tty = isatty(0);
    out_is_tty = isatty(1);
    on_tty = in_is_tty && out_is_tty;
    setlocale(LC_CTYPE, "");
    init_iconv();
    buffer = NULL;
    buffer_size = 0;
    buffer_have = 0;
    connect_jtalkd();
    send_argv(argc - 1, argv + 1);
    rl_callback_handler_install(on_tty ? SAVE_CURSOR_PROMPT ">" : "",
	have_keyboard_line);
    fds[0].fd = sock;
    fds[0].events = POLLIN;
    fds[1].fd = 0;
    fds[1].events = POLLIN;
    in_enabled = 1;
    while(1) {
	if(poll(fds, 1 + in_enabled, -1) < 0) {
	    if(errno == EINTR)
		continue;
	    perror("poll");
	    exit(1);
	}
	if((fds[0].revents & POLLIN) != 0)
	    read_from_jtalkd();
	if((fds[1].revents & POLLIN) != 0)
	    rl_callback_read_char();
    }
    return(0);
}
